/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [scn_io.c]             Scanner Definition Primimitives                   */
/*                                                                          */
/* Copyright (c) 1993 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

/* TODO                                                                     */
/*                                                                          */
/* - Add a Scn_usageEof together with a character set size                  */
/* - cdefeof --> cdeftok                                                    */

#include "standard.h"

#include "scn_imp.h"
#include "pathes.h"
#include "binimg.h"
#include "sink.h"

/*!!
   Beside copying and disposing of scanner definitions 
   this module provides functions that 
   1) generate a C module for a given scanner definition
   2) unload a scanner definition to a binary image file 
   3) load a scanner definition from a binary image file.
*/

/*I------------------------------ Types & Macros ---------------------------- */

/*!
  <p><b>Identifier of other other</b>
*/

/*DOC_INTERFACE*/
#define SCN_TOK_Other "[other]"

AbstractType( Scn_T ); /* Abstract scanner type */

/*I------------------------------ Basics ------------------------------------ */

string Scn_LanguageId(Scn_T x)
/* language name of scanner definition 'x' */
{
  return Scn_IMP(x)->Name;
}

Scn_T Scn_copy(Scn_T x)
/* copies scanner definition 'x' */
{ scn_t scn = Scn_IMP(x); int i;
  scn_t res;
  res = New(scn_t);
  res->Name    = StrCopy(scn->Name);
  res->States  = scn->States;
  res->Tokens  = scn->Tokens;
  res->StaEdg  = (short  *)BytCopy(scn->StaEdg,(scn->States+1) * sizeof(short  ));
  res->StaFin  = (short  *)BytCopy(scn->StaFin,scn->States     * sizeof(short  ));
  res->EdgeC   = (wc_char*)BytCopy(scn->EdgeC, SCN_Edges(scn)  * sizeof(wc_char));
  res->EdgeS   = (short  *)BytCopy(scn->EdgeS, SCN_Edges(scn)  * sizeof(short  ));
  res->TokId   = (string *)NewMem(             scn->Tokens     * sizeof(string ));
  for(i = 0; i < scn->Tokens; i++)
    res->TokId[i] = StrCopy(scn->TokId[i]);
  res->Flags   = (byte*)BytCopy(scn->Flags,  scn->Tokens     * sizeof(byte  ));
  return Scn_ADT(res);
}

void Scn_free(Scn_T x)
/* frees scanner definition 'x' */
{ scn_t scn = Scn_IMP(x); int i;
  FreeMem(scn->Name);
  FreeMem(scn->StaEdg);
  FreeMem(scn->StaFin);
  FreeMem(scn->EdgeC);
  FreeMem(scn->EdgeS);
  for(i = 0; i < scn->Tokens; i++)
    FreeMem(scn->TokId[i]);
  FreeMem(scn->TokId);
  FreeMem(scn->Flags);
  FreeMem(scn);
}

/*I----------------------------- Binary Image ------------------------------- */

Scn_T Scn_get_bin(void)
/* loads a scanner definition 
   from the current ( previously opened ) image
*/
{ int i; scn_t scn;
  scn = New(scn_t);
  getString(&scn->Name);
  getWord(&scn->States);
  getWord(&scn->Tokens);
  scn->StaEdg = (short*)NewMem( (scn->States+1) * sizeof(short) );
  for (i = 0; i < scn->States+1; i++) getWord(&scn->StaEdg[i]);
  scn->StaFin = (short*)NewMem( scn->States * sizeof(short) );
  for (i = 0; i < scn->States; i++) getWord(&scn->StaFin[i]);
  scn->EdgeC = (wc_char*)NewMem( sizeof(wc_char )  * SCN_Edges(scn) );
  for (i = 0; i < SCN_Edges(scn); i++)
  { unsigned long ul;
    getULong(&ul);
    scn->EdgeC[i] = (wc_char)ul;
  }
  scn->EdgeS = (short*)NewMem( sizeof(short) * SCN_Edges(scn) );
  for (i = 0; i < SCN_Edges(scn); i++) getWord(&scn->EdgeS[i]);
  scn->TokId = (string*)NewMem( sizeof(string) * scn->Tokens );
  for (i = 0; i < scn->Tokens; i++) getString(&scn->TokId[i]);
  scn->Flags = (byte*)NewMem( sizeof(byte) * scn->Tokens );
  for (i = 0; i < scn->Tokens; i++) getByte(&scn->Flags[i]);
  return scn;
}

void Scn_get(string Language, Scn_T *x)
/* loads a scanner definition named 'Language' into 'x'
   from image [$PATH/]<Language>.lim
*/
{
  getBgn("PATH",Language,".lim");
  getHeader("lim",1,0);
  *x = Scn_get_bin();
  getEnd();
}

void Scn_put_bin(Scn_T x)
/* unloads scanner definition 'x'
   to the current ( previously opened ) image
*/
{ int i; scn_t scn = Scn_IMP(x);
  putString(scn->Name);
  putWord(scn->States);
  putWord(scn->Tokens);
  for (i = 0; i < scn->States+1 ; i++) putWord  (scn->StaEdg[i]);
  for (i = 0; i < scn->States   ; i++) putWord  (scn->StaFin[i]);
  for (i = 0; i < SCN_Edges(scn); i++) putULong ((unsigned long)scn->EdgeC [i]);
  for (i = 0; i < SCN_Edges(scn); i++) putWord  (scn->EdgeS [i]);
  for (i = 0; i < scn->Tokens   ; i++) putString(scn->TokId [i]);
  for (i = 0; i < scn->Tokens   ; i++) putByte  (scn->Flags[i]);
}

void Scn_put(c_string Environ, Scn_T x)
/* unloads scanner definition 'x'
   to image [$'Environ'/]<Language>.lim
*/
{ string hd; scn_t scn = Scn_IMP(x);
  putBgn(Environ,scn->Name,".lim");
  hd = Str_printf
       (
         "[%s.lim] Binary scanner image for '%s'\n", scn->Name, scn->Name
       );
  putHeader(hd,"lim",1,0);
  Scn_put_bin(x);
  FreeMem(hd);
  putEnd();
}

/*I----------------------------- C-Module ----------------------------------- */

void Scn_C_Modul(c_string Environ, Scn_T x)
/* creates source [$'Environ'/]<Language>_lim.c 
   for scanner definition 'x'
*/
{ FILE *tf; int i; scn_t scn = Scn_IMP(x); char fid[9];
  sprintf(fid,"%.4s_lim",scn->Name);
  tf = OpenPath(Environ,fid,".c","wt");
  fprintf(tf,"/* [%s.c] Scanner table for '%s'  */\n\n", fid, scn->Name);
  fprintf(tf,"#include \"scn_io.h\"\n\n");
  fprintf(tf,"#include \"scn_imp.h\"\n\n");
  fprintf(tf,"static short StaEdg[%d] =\n",scn->States+1);
  fprintf(tf,"/* first edge per state */\n");
  fprintf(tf,"{\n  ");
  for (i = 0; i < scn->States+1; i++)
  {
    fprintf(tf,"%3d",scn->StaEdg[i]);
    if (             i+1 < scn->States+1) fprintf(tf,",");
    if (i%10 != 9 && i+1 < scn->States+1) fprintf(tf," ");
    if (i%10 == 9 && i+1 < scn->States+1) fprintf(tf,"\n  ");
  }
  fprintf(tf,"\n};\n\n");
  fprintf(tf,"static short StaFin[%d] =\n",scn->States);
  fprintf(tf,"/* finite token per state */\n");
  fprintf(tf,"{\n  ");
  for (i = 0; i < scn->States; i++)
  {
    fprintf(tf,"%3d",scn->StaFin[i]);
    if (             i+1 < scn->States) fprintf(tf,",");
    if (i%10 != 9 && i+1 < scn->States) fprintf(tf," ");
    if (i%10 == 9 && i+1 < scn->States) fprintf(tf,"\n  ");
  }
  fprintf(tf,"\n};\n\n");
  fprintf(tf,"static wc_char EdgeC[%d] =\n",SCN_Edges(scn));
  fprintf(tf,"/* lowest character per edge */\n");
  fprintf(tf,"{ /*State*/\n");
  for (i = 0; i < scn->States; i++)
  { int j, ed = scn->StaEdg[i+1] - scn->StaEdg[i];
    fprintf(tf,"  /* %3d */", i);
    for (j = 0; j < ed; j++)
    { int bgn = scn->StaEdg[i];
      fprintf(tf,"  %9ld",scn->EdgeC[bgn+j]);
      if (             bgn+j+1 < SCN_Edges(scn)) fprintf(tf,",");
      if (j%8 == 7 && bgn+j+1 < SCN_Edges(scn) && j+1 < ed)
        fprintf(tf,"\n           ");
    }
    fprintf(tf,"\n");
  }
  fprintf(tf,"};\n\n");
  fprintf(tf,"static short EdgeS[%d] =\n",SCN_Edges(scn));
  fprintf(tf,"/* follow (state+1) per edge */\n");
  fprintf(tf,"{ /*State*/\n");
  for (i = 0; i < scn->States; i++)
  { int j, ed = scn->StaEdg[i+1] - scn->StaEdg[i];
    fprintf(tf,"  /* %3d */", i);
    for (j = 0; j < ed; j++)
    { int bgn = scn->StaEdg[i];
      fprintf(tf," %3d",scn->EdgeS[bgn+j]);
      if (             bgn+j+1 < SCN_Edges(scn)) fprintf(tf,",");
      if (j%10 == 9 && bgn+j+1 < SCN_Edges(scn) && j+1 < ed)
        fprintf(tf,"\n           ");
    }
    fprintf(tf,"\n");
  }
  fprintf(tf,"};\n\n");
  fprintf(tf,"static string Tokens[%d] =\n", scn->Tokens);
  fprintf(tf,"/* token -> string */\n");
  fprintf(tf,"{\n");
  for (i = 0; i < scn->Tokens; i++)
    fprintf(tf,"  \"%s\"%s\n",scn->TokId[i],i+1<scn->Tokens?",":"");
  fprintf(tf,"};\n\n");
  fprintf(tf,"static byte Flags[%d] =\n", scn->Tokens);
  fprintf(tf,"/* token -> newid */\n");
  fprintf(tf,"{\n  ");
  for (i = 0; i < scn->Tokens; i++)
  {
    fprintf(tf,"0x%02x", scn->Flags[i]);
    if (i+1 < scn->Tokens)                fprintf(tf,", ");
    if (i%10 == 9 && (i+1 < scn->Tokens)) fprintf(tf,"\n  ");
  }
  fprintf(tf,"\n};\n\n");
  fprintf(tf,"static ConcreteImp(scn_t) Scanner =\n");
  fprintf(tf,"/* the scanner structure */\n");
  fprintf(tf,"{\n");
  fprintf(tf,"  \"%s\", /* Name    */\n",scn->Name   );
  fprintf(tf,"  %2d,%*s /* States  */\n",scn->States , strlen(scn->Name), "");
  fprintf(tf,"  %2d,%*s /* Tokens  */\n",scn->Tokens , strlen(scn->Name), "");
  fprintf(tf,"  StaEdg,\n");
  fprintf(tf,"  StaFin,\n");
  fprintf(tf,"  EdgeC,\n");
  fprintf(tf,"  EdgeS,\n");
  fprintf(tf,"  Tokens,\n");
  fprintf(tf,"  Flags,\n");
  fprintf(tf,"};\n\n");
  fprintf(tf,"void Scn_get_%s(Scn_T *scn)\n",scn->Name);
  fprintf(tf,"/* simulates a Scn_get */\n");
  fprintf(tf,"{\n");
  fprintf(tf,"  *scn = Scn_copy(Scn_ADT(&Scanner));\n");
  fprintf(tf,"};\n");
  fclose(tf);
}
