/* ------------------------------------------------------------------------ */
/*                                                                          */
/* [lim_test.c]                  Scanner Test                               */
/*                                                                          */
/* Copyright (c) 1993 by D\olle, Manns                                      */
/* ------------------------------------------------------------------------ */

#include "stdosx.h"
#include "scn_base.h"
#include "gstream.h"
#include "pathes.h"

#include "lim__cim.h"

int main(int argc, string argv[])
{ Scn_Stream stream; Scn_T scn; int i, maxlen;
  string cset, file;

  CTX_init_lim_test(argc,argv);
  initSymbols();
  cset = CTX_EVAL(string,"charset");
  file = CTX_EVAL(string,"FileName");
  /* get the scanner */
  Scn_get(CTX_EVAL(string,"Language"), &scn);
  /* prepare layout */
  maxlen = 0;
  for (i = 0; i < Scn_tokens(scn); i++)
  { string s = Scn_tokid(scn,i);
    maxlen = MAX(maxlen, strlen(s));
    FreeMem(s);
  }
  /* actual scanning */
  if( strlen(cset) == 0 )
    stream = Stream_file(scn, "", file, "");
  else
  { FILE*          fp  = OpnFile(file,"rb");
    GStream_T      gs  = GS_stream_file_new(fp, cset, True); 
    Scn_Stream_Itr itr;
    assert1(gs != (GStream_T)NULL, "creation of stream '%s' failed",file);
    itr = Stream_Itr_new
          (
            GS_stream_get_wcrc, GS_fun_destruct(gs), GS_ucs4_to_utf8, 
            GS_fun_seek(gs), gs, file
          );
    stream = Stream_bgn(scn,itr);
    Stream_Itr_free(itr);
  }
  Stream_defEofId(stream,-1);
  Stream_defErrId(stream, 0);
  for (i = 1; i < Scn_tokens(scn); i++)
  { string s = Scn_tokid(scn,i);
    Stream_defTokId( stream, s, (short)i );
    FreeMem(s);
  }
  for
  (
    Stream_next(stream);
    Stream_ctid(stream) >= 0;
    Stream_next(stream)
  )
  if (CTX_EVAL(bool,"tokens"))
  { string s = Scn_tokid(scn,Stream_ctid(stream));
    printf
    (
      "%s:%03d:%03d %s%*s : ",
      symbolToString(Stream_cfil(stream)),
      Stream_clin(stream),
      Stream_ccol(stream),
      s,
      maxlen - strlen(s), ""
    );
    if( strlen(cset) == 0 )
      printf("%s\n", symbolToString(Stream_csym(stream)));
    else
    {
      if( CTX_EVAL(bool,"wprint") )
        GS_fprint_utf8(STDOUT,symbolToString(Stream_csym(stream)),False);
      else GS_fprint_utf8(STDOUT,symbolToString(Stream_csym(stream)),True);
      printf("\n");
    }
    FreeMem(s);
  }
  Stream_close(stream);
  Stream_free(stream);
  Scn_free(scn);
  if (CTX_EVAL(bool,"symbols")) { NL; printSymbols(0); NL; }
  freeSymbols();
  CTX_quit();
  BUG_CORE;
  return(0);
}
